import dataclasses

from controllogger.misc.base_logger_config import BaseLoggerConfig
from controllogger.misc.logger_defaults_config import LoggerDefaultsConfig


@dataclasses.dataclass
class InputLoggerConfig(BaseLoggerConfig):
    def __default_init__(self, logger_defaults_config: LoggerDefaultsConfig):
        super().__default_init__(logger_defaults_config)



import dataclasses
from pathlib import Path

from controllogger.enums.log_levels import LogLevels
from controllogger.enums.terminal_out_file import TerminalOutFile
from controllogger.misc.base_dataclass import BaseDataclass


@dataclasses.dataclass
class LoggerDefaultsConfig(BaseDataclass):
    name: str = None  # Used as base name for output loggers.
    level: LogLevels | int = LogLevels.NOTSET  # Default log level for output loggers. Default is NOTSET.
    last_resort: bool = False  # Default last resort for output loggers. Last resort is used when the output logger is not able to handle the log record.
    log_events: bool = False  # Default log events for output loggers. Is used to log events to loggers. Default is False.
    log_init: bool | None = None  # Default log init for input and output loggers. Is used to log the init event to loggers. Default is False.
    log_attach: bool | None = None  # Default log attach for input and output loggers. Is used to log the attach event to loggers. Default is False.
    log_detach: bool | None = None  # Default log detach for input and output loggers. Is used to log the detach event to loggers. Default is False.
    log_destroy: bool | None = None  # Default log destroy for input and output loggers. Is used to log the destroy event to loggers. Default is False.

    # Default values for output logger
    header_width: int = 110  # Default header width. Default is 110.
    console: bool = False  # Default console. Default is False.
    console_format: str = "[%(asctime)s][%(name)s][%(levelname)s] %(srcRelativePathname)s(%(srcLineno)s) -> %(relativePathname)s(%(lineno)s) - %(message)s"  # Default console format. Default is %(name)s\t - %(message)s.
    console_outfile: TerminalOutFile = TerminalOutFile.STDOUT  # Default console outfile. Default is STDOUT.
    console_width: int = 160  # Default console width. Default is 160(my preferred width with 49 inch monitor ^^).
    console_rich_force_terminal: bool = False  # Default console rich force terminal. Default is False. Only used if rich is installed.
    file_path: Path | str = Path("logs")  # Base path for used by file output loggers. If a relative path is given, it will be relative to the current working directory.
    file_format: str = "%(asctime)s - %(levelname)s - %(name)s - %(message)s"  # Default file format. Default is %(asctime)s - %(levelname)s - %(name)s - %(message)s.
    file_mode: str = "a"  # Default file mode. Default is a.
    file_max_size: int = 1024 * 1024 * 10  # Default file max size. Default is 10 MB.
    file_backup_count: int = 0  # Default file backup count. Default is 0.
    file_encoding: str = "utf-8"  # Default file encoding. Default is utf-8.
    file_delay: bool = False  # Default file delay. Default is False.
    file_archive_backup_count: int = 0  # Default file archive backup count. Default is 0.
    db_buffer_max_lines: int = 100  # Default db buffer max lines. Default is 100.
    db_buffer_max_time: int = 10  # Default db buffer max time. Default is 10.



from datetime import datetime
from enum import Enum
from pathlib import Path

from pydantic import Field, BaseModel

from ..classes.pydantic_object_id import PydanticObjectId
from ..classes.singleton import Singleton
from ..enums.log import LogLevels
from ..enums.terminal_out_file import TerminalOutFile





class LogEntryModel(BaseModel):
    """
    Class for log entries
    """

    class Config:  # Shadows models.base.BaseModel.Config
        use_enum_values = True

    timestamp: datetime = Field(..., title="Timestamp", description="Timestamp of log entry")
    level: LogLevels = Field(..., title="Level", description="Level of log entry")
    msg: str = Field(..., title="Message", description="Message of log entry")


class BaseLogNewModel(BaseModel):
    """
    Base Class for new log
    """

    class Config:  # Shadows models.base.BaseModel.Config
        use_enum_values = True

    id: str = Field(..., title="ID", description="ID of background task")
    name: str = Field(..., title="Name", description="Name of background task")


class BaseLogReadModel(BaseModel):
    """
    Base Class for read log
    """

    class Config:  # Shadows models.base.BaseModel.Config
        use_enum_values = True

    # Shadows models.base.BaseDbReadModel:
    id: PydanticObjectId = Field(..., title="ID", description="ID of document")
    created_at: float = Field(..., title="Created At", description="Created at timestamp")
    updated_at: float = Field(..., title="Updated At", description="Updated at timestamp")
    # End Shadows

    name: str = Field(..., title="Name", description="Name of log")
    started_at: datetime = Field(None, title="Started At", description="Timestamp of when log started")
    finished_at: datetime = Field(None, title="Finished At", description="Timestamp of when log finished")
    log_lines: list[LogEntryModel] = Field([], title="Log Lines", description="List of log lines")


class BaseLogListModel(BaseModel):
    """
    Base Class for list log
    """

    class Config:  # Shadows models.base.BaseModel.Config
        use_enum_values = True

    # Shadows models.base.BaseDbReadModel:
    id: PydanticObjectId = Field(..., title="ID", description="ID of document")
    created_at: float = Field(..., title="Created At", description="Created at timestamp")
    updated_at: float = Field(..., title="Updated At", description="Updated at timestamp")
    # End Shadows

    name: str = Field(..., title="Name", description="Name of background task")
    started_at: datetime = Field(None, title="Started At", description="Started at")
    finished_at: datetime = Field(None, title="Finished At", description="Finished at")


class BaseLogWriteModel(BaseModel):
    """
    Base Class for write log
    """

    class Config:  # Shadows models.base.BaseModel.Config
        use_enum_values = True

    name: str = Field(..., title="Name", description="Name of log")
    started_at: datetime = Field(None, title="Started At", description="Timestamp of when log started")
    finished_at: datetime = Field(None, title="Finished At", description="Timestamp of when log finished")
    log_lines: list[LogEntryModel] = Field([], title="Log Lines", description="List of log lines")



import dataclasses
import logging
from abc import ABC
from pathlib import Path
from typing import Callable

import controllogger


class BaseEasyLogger(logging.Logger, ABC):

    @staticmethod
    def _pars_relative_path(path: Path | str, relative_path: Path | str) -> tuple[bool, str]:
        try:
            relative_path_name = str(Path(path).relative_to(relative_path))
        except ValueError:
            return False, path
        return True, relative_path_name

    def makeRecord(self, *args, **kwargs):
        record = super().makeRecord(*args, **kwargs)

        i = 1
        previous_frame = None
        while True:
            frame = self.findCaller(stacklevel=i)
            if previous_frame is not None:
                if previous_frame == frame:
                    break
            if controllogger.__script_start_file__ in frame[0]:
                break
            previous_frame = frame
            i += 1
        _, src_relative_path_name = self._pars_relative_path(frame[0], controllogger.__script_start_file__)
        result, relative_path_name = self._pars_relative_path(path=record.pathname, relative_path=controllogger.__script_start_file__)
        if not result:
            result, relative_path_name = self._pars_relative_path(path=record.pathname, relative_path=controllogger.__module_path__)
            if not result:
                result, relative_path_name = self._pars_relative_path(path=record.pathname, relative_path=controllogger.__lib_path__)
                if not result:
                    result, relative_path_name = self._pars_relative_path(path=record.pathname, relative_path=controllogger.__site_packages_path__)
                    if not result:
                        relative_path_name = record.pathname

        for key, value in {
            "relativePathname": relative_path_name,
            "srcPathname": frame[0],
            "srcRelativePathname": src_relative_path_name,
            "srcFilename": Path(frame[0]).name,
            "srcLineno": frame[1],
            "srcFunctionName": frame[2],
            "srcStackLevel": i
        }.items():
            if hasattr(record, key):
                raise ValueError(f"Record already has attribute '{key}'.")
            setattr(record, key, value)
        return record



import dataclasses
from pathlib import Path

from controllogger.enums.terminal_out_file import TerminalOutFile
from controllogger.misc.base_logger_config import BaseLoggerConfig
from controllogger.misc.logger_defaults_config import LoggerDefaultsConfig
from controllogger.misc.filter import Filter


@dataclasses.dataclass
class OutputLoggerConfig(BaseLoggerConfig):
    filter: list[Filter] = dataclasses.field(default_factory=list)  # List of filters
    last_resort: bool | None = None  # Last resort of output logger
    header_width: int | None = None  # Header width of output logger
    console: bool = False  # Enable console output
    console_format: str | None = None  # Console format of output logger
    console_outfile: TerminalOutFile | None = None  # Console outfile of output logger
    console_width: int | None = None  # Console width of output logger
    console_rich_force_terminal: bool | None = None  # Default console rich force terminal. Only used if rich is installed.
    file: bool = False  # Enable file output
    file_format: str | None = None  # File format of output logger
    file_path: Path | str | None = None  # File path of output logger
    file_mode: str | None = None  # File mode of output logger
    file_max_size: int | None = None  # File max size of output logger
    file_backup_count: int | None = None  # File backup count of output logger
    file_encoding: str | None = None  # File encoding of output logger
    file_delay: bool | None = None  # File delay of output logger
    file_archive_backup_count: int | None = None  # File archive backup count of output logger
    db: bool = False  # Enable db output
    db_format: str | None = None  # DB format of output logger
    db_buffer_max_lines: int | None = None  # DB buffer max lines of output logger
    db_buffer_max_time: int | None = None  # DB buffer max time of output logger

    def __default_init__(self, logger_defaults_config: LoggerDefaultsConfig):
        super().__default_init__(logger_defaults_config)
        # setup None values

        # last_resort
        if self.last_resort is None:
            self.last_resort = logger_defaults_config.last_resort

        # header_width
        if self.header_width is None:
            self.header_width = logger_defaults_config.header_width

        # console
        if self.console:
            # console_format
            if self.console_format is None:
                self.console_format = logger_defaults_config.console_format
            # console_outfile
            if self.console_outfile is None:
                self.console_outfile = logger_defaults_config.console_outfile
            # console_width
            if self.console_width is None:
                self.console_width = logger_defaults_config.console_width
            # console_rich_force_terminal
            if self.console_rich_force_terminal is None:
                self.console_rich_force_terminal = logger_defaults_config.console_rich_force_terminal

        # file
        if self.file:
            # file_format
            if self.file_format is None:
                self.file_format = logger_defaults_config.file_format
            # file_path
            if self.file_path is None:
                self.file_path = Path(logger_defaults_config.file_path) / f"{self.name}.log"
            # file_mode
            if self.file_mode is None:
                self.file_mode = logger_defaults_config.file_mode
            # file_max_size
            if self.file_max_size is None:
                self.file_max_size = logger_defaults_config.file_max_size
            # file_backup_count
            if self.file_backup_count is None:
                self.file_backup_count = logger_defaults_config.file_backup_count
            # file_encoding
            if self.file_encoding is None:
                self.file_encoding = logger_defaults_config.file_encoding
            # file_delay
            if self.file_delay is None:
                self.file_delay = logger_defaults_config.file_delay
            # file_archive_backup_count
            if self.file_archive_backup_count is None:
                self.file_archive_backup_count = logger_defaults_config.file_archive_backup_count

        # db
        if self.db:
            # db_format
            if self.db_format is None:
                self.db_format = logger_defaults_config.db_format
            # db_buffer_max_lines
            if self.db_buffer_max_lines is None:
                self.db_buffer_max_lines = logger_defaults_config.db_buffer_max_lines
            # db_buffer_max_time
            if self.db_buffer_max_time is None:
                self.db_buffer_max_time = logger_defaults_config.db_buffer_max_time



import logging

import io
import threading

from datetime import datetime

from controllogger.handler.base import BaseLogHandler


class DbLogHandler(logging.StreamHandler, BaseLogHandler):
    def __init__(
            self,
            log_document: BaseLogDocument,
            buffer_max_lines: int = 10,
            buffer_max_time: int = 60,
    ):
        super().__init__(stream=io.StringIO())

        self.log_document = log_document

        # check if log_document is a instance of BaseLogDocument
        if not isinstance(self.log_document, BaseLogDocument):
            raise TypeError(f"Expected log_document to be a instance of BaseLogDocument, got {type(self.log_document)}.")

        self.buffer_len = 0
        self.buffer_max_lines = buffer_max_lines
        self.buffer_max_time = buffer_max_time
        self.max_time_callback_timer = threading.Timer(self.buffer_max_time, self._max_time_callback)

    def close(self):
        super().close()

        # flush buffer
        if self.buffer_len > 0:
            self.flush()

    def emit(self, record: logging.LogRecord) -> None:
        # get log level
        level = dict(LogLevels.__members__)[record.levelname]
        msg = self.format(record)

        # create log line
        log_line = LogEntryModel(timestamp=datetime.now(), level=level, msg=msg)
        log_line_str = log_line.json()

        # add log line to stream
        self.stream.write(log_line_str + "\n")
        self.buffer_len += 1

        # flush buffer if full
        if self.buffer_len >= self.buffer_max_lines:
            self.flush()
        else:
            # setup timed callback to flush buffer
            if not self.max_time_callback_timer.is_alive():
                root_logger.debug(f"Setting up timed callback for '{self.log_document.name}'.")
                self.max_time_callback_timer = threading.Timer(self.buffer_max_time, self._max_time_callback)
                self.max_time_callback_timer.start()

    def _max_time_callback(self) -> None:
        if not self.max_time_callback_timer.is_alive():
            raise RuntimeError("Callback is not alive.")
        root_logger.debug(f"Callback timed out for '{self.log_document.name}'. -> Flushing buffer.")
        self.flush()

    def flush(self) -> None:
        # reset timed callback
        if self.max_time_callback_timer.is_alive():
            root_logger.debug(f"Resetting timed callback for '{self.log_document.name}'.")
            self.max_time_callback_timer.cancel()

        # check if buffer is empty
        if self.buffer_len == 0:
            root_logger.debug(f"Buffer is empty for '{self.log_document.name}'.")
            return

        root_logger.debug(f"Flushing buffer for '{self.log_document.name}'.")

        for buffer_line in self.buffer:
            line = LogEntryDocument(**buffer_line.dict())
            self.log_document.log_lines.append(line)
        self.reset_buffer()

        self.log_document.save()
        root_logger.debug(f"Buffer flushed for '{self.log_document.name}'.")

    def reset_buffer(self):
        root_logger.debug(f"Resetting buffer for '{self.log_document.name}'.")
        self.stream.seek(0)
        self.stream.truncate()
        self.buffer_len = 0

    @property
    def buffer(self) -> list[LogEntryModel]:
        stream_str = self.stream.getvalue()
        buffer_lines = []
        for line_str in stream_str.split("\n"):
            if line_str.strip() == "":
                continue
            buffer_lines.append(LogEntryModel.parse_raw(line_str))
        return buffer_lines



import logging

from controllogger.enums.log_levels import LogLevels
from controllogger.logger.base import BaseLogger
from controllogger.misc.input_logger_config import InputLoggerConfig


class InputLogger(BaseLogger):
    _ConfigClass = InputLoggerConfig
    _my_type: str = "InputLogger"
    _control_logger_type_field = "input_logger_type"

    def __init__(self, name: str = None, level: LogLevels | int = logging.NOTSET, config: _ConfigClass | dict[str, any] = None):
        super().__init__(name=name, level=level, config=config)

        self.attach()

    def __del__(self):
        if not self.open:
            return
        super().__del__()
        self.detach()
        del self.control_logger.all_logger[self.name]

    def __str__(self) -> str:
        name = self.name
        level = LogLevels(self.level).name
        context_str = ", ".join([str(context) for context in self.contexts])
        r = f"{self.__class__.__name__}({name=}, {level=}, "
        if self.closed:
            r = r[:-2]
            r += ") CLOSED"
            return r
        if not self.attached:
            r = r[:-2]
            r += ") DETACHED"
            return r
        if len(self.contexts) > 0:
            r += f"contexts=[{context_str}], "
        r = r[:-2]
        r += ")"

        return r

    def attach(self) -> None:
        """
        Attach the input logger to the control logger
        :return: None
        """

        if self.attached:
            return None

        if self.name in self.control_logger.input_loggers.keys():
            raise ValueError(f"Input logger for '{self.name}' already exists.")
        self.control_logger.input_loggers[self.name] = self
        super().attach()

    def detach(self) -> None:
        """
        Detach the input logger from the control logger
        :return: None
        """

        if not self.attached:
            return None

        if self.name not in self.control_logger.input_loggers.keys():
            raise ValueError(f"Input logger for '{self.name}' not found.")
        super().detach()
        del self.control_logger.input_loggers[self.name]

    @property
    def attached(self) -> bool:
        for logger_name in self.control_logger.input_loggers.keys():
            if logger_name == self.name:
                return True
        return False

    def reconfigure(self):
        self.setLevel(self.control_logger.level)
        self.propagate = self.control_logger.propagate
        self.disabled = self.control_logger.disabled

    def makeRecord(self, *args, **kwargs):
        record = super().makeRecord(*args, **kwargs)

        # set context
        record.contexts = self.contexts

        return record

    def handle(self, record):
        self._skip_handle = True
        super().handle(record)
        # handle record by control logger
        return self.control_logger.handle(record)

    def isEnabledFor(self, level):
        self.reconfigure()
        return super().isEnabledFor(level)



import dataclasses

from controllogger.enums.log_levels import LogLevels
from controllogger.logger.input import InputLogger
from controllogger.logger.output import OutputLogger
from controllogger.misc.logger_defaults_config import LoggerDefaultsConfig
from controllogger.misc.output_logger_config import OutputLoggerConfig


@dataclasses.dataclass
class EasyLoggerConfig(LoggerDefaultsConfig):
    name: str = "easy_logger"  # Name of the control logger and also used as base name for output loggers.
    level: LogLevels | int = LogLevels.DEBUG  # Log level for the control logger. All input loggers will automatically use this level.
    last_resort: bool = False  # Default last resort for the control logger.
    # Last resort is used when no output logger is found for a log record or the output logger is not able to handle the log record.

    output_loggers: list[OutputLoggerConfig] = dataclasses.field(default_factory=lambda: [OutputLoggerConfig(name="console", console=True),
                                                                                          OutputLoggerConfig(name="file",
                                                                                                             file=True)])
    # List of output loggers created by the control logger on startup.

    # define base classes
    input_logger_cls: type[InputLogger] = InputLogger
    output_logger_cls: type[OutputLogger] = OutputLogger



from typing import NewType, Callable

from controllogger.logger.input import InputLogger
from controllogger.misc.context import LoggerContext

LoggerContextImplementation = NewType("LoggerContextImplementation", LoggerContext) | Callable
InputLoggerImplementation = NewType("InputLoggerImplementation", InputLogger)


class LoggerClass:
    logger_context: LoggerContextImplementation
    logger: InputLoggerImplementation



from rich.logging import RichHandler

from kdsm_api.classes.logging.base import BaseLogHandler
from kdsm_api.enums.terminal_out_file import TerminalOutFile
from kdsm_api.functions.get_rich_console import get_rich_console


class RichConsoleHandler(RichHandler, BaseLogHandler):
    def __init__(self, console_outfile: TerminalOutFile, console_width: int, console_force_terminal: bool):
        super().__init__(
            console=get_rich_console(out_file=console_outfile,
                                     width=console_width,
                                     force_terminal=console_force_terminal),
        )

    # def handle(self, record):
    #     if record.msg == "[HEADER]" and hasattr(record, "header"):
    #         print()
    #     super().handle(record)



import logging

last_resort = logging.getLogger("fastapi-easylogger-last-resort")
last_resort_handler = logging.StreamHandler()
last_resort_handler_formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
last_resort_handler.setFormatter(last_resort_handler_formatter)
last_resort.addHandler(last_resort_handler)


from controllogger.enums.operator import Operator
from controllogger.misc.base_dataclass import BaseDataclass


class Filter(BaseDataclass):
    key: str
    value: bool | int | float | str | list[bool | int | float | str] | dict[str, bool | int | float | str]
    operator: Operator = Operator.EQUALS



import subprocess
import sys
from pathlib import Path

import codecs
import os.path


def setup_exit(exit_code: int = 0, message: str = None):
    if message is not None:
        print(message)
    raise SystemExit(exit_code)


def _read(rel_path: Path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()


def get_version(rel_path: str | Path) -> str:
    if isinstance(rel_path, str):
        rel_path = Path(rel_path)
    for line in _read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            v = line.split(delim)[1]
            break
    else:
        raise RuntimeError("Unable to find version string.")
    if build_version:
        v += f".{build_version}"
    return v


def print_version(rel_path: str | Path):
    print(f"Current version: {get_version(rel_path)}")
    setup_exit(0)


def _build():
    # run sdist
    cmd = ["python", "setup.py"]
    if build_version:
        cmd.append("--build-version")
        cmd.append(build_version)
    cmd.append("sdist")
    if subprocess.run(cmd).returncode != 0:
        setup_exit(1, "Building sdist failed.")


def _setup():
    from setuptools import setup
    setup(version=get_version("src/controllogger/__init__.py"))


if __name__ == "__main__":
    if "-v" in sys.argv or "--version" in sys.argv:
        print_version("src/controllogger/__init__.py")

    build_version = None
    if "-bV" in sys.argv or "--build-version" in sys.argv:
        index = sys.argv.index("-bV" if "-bV" in sys.argv else "--build-version") + 1
        if index >= len(sys.argv):
            setup_exit(1, "Missing version number.")
        if sys.argv[index].startswith("-"):
            setup_exit(1, "Missing version number.")
        build_version = sys.argv[index]
        sys.argv.pop(index)
        sys.argv.pop(index - 1)

    if "build" in sys.argv:
        _build()
    else:
        _setup()



import dataclasses

from controllogger.enums.log_levels import LogLevels
from controllogger.misc.base_dataclass import BaseDataclass


@dataclasses.dataclass
class LogFuntionConfig(BaseDataclass):
    header: bool = True  # Log function header. Default is True.
    footer: bool = True  # Log function footer. Default is True.
    header_footer_level: LogLevels = LogLevels.DEBUG  # Log function header and footer level. Default is DEBUG.
    start_msg: str | None = None  # Log function start message. Default is None.
    end_msg: str | None = None  # Log function end message. Default is None.
    start_end_msg_level: LogLevels = LogLevels.INFO  # Log function start message level. Default is DEBUG.
    measure_time: bool = False  # Log function measure time. Default is False.



import os
from pathlib import Path

import setuptools

__author__ = "Julius Koenig"
__version__ = "0.2"

# overwrite version if "__dev_version__" is set in environment
if "__dev_version__" in os.environ:
    __version__ = os.environ["__dev_version__"]

__script_start_file__ = os.getcwd()
__site_packages_path__ = str(Path(setuptools.__file__).parent.parent)
__lib_path__ = str(Path(os.__file__).parent)
__module_path__ = os.path.dirname(os.path.abspath(__file__))
__module_files__ = []



from abc import ABCMeta
from typing import Any


class Singleton(ABCMeta):
    """
    Singleton metaclass
    """

    singleton_map: dict[str, Any] = {}

    def __new__(cls, name, bases, attrs):
        __init__ = attrs.pop("__init__", None)

        def singleton__init__(self, *args, **kwargs):
            self_name = self.__class__.__name__
            if self_name not in Singleton.singleton_map:
                Singleton.singleton_map[self_name] = self
                if __init__ is not None:
                    __init__(self, *args, **kwargs)
            else:
                raise RuntimeError(f"Singleton {self_name} already initialized. Use {self_name}() to get the instance.")

        attrs["__singleton__"] = True

        # check if __singleton__ is set on bases
        singleton_set = False
        for base in bases:
            if hasattr(base, "__singleton__"):
                singleton_set = True
                break

        if not singleton_set:
            attrs["__init__"] = singleton__init__

        # check if BaseModel in bases
        base_model_set = False
        for base in bases:
            if base.__name__ == "BaseModel":
                base_model_set = True
                break

        if not base_model_set:
            return ABCMeta.__new__(cls, name, bases, attrs)
        else:
            return super().__new__(cls, name, bases, attrs)

    def __call__(cls, *args, init: bool = True, **kwargs):
        if init:
            if cls.__name__ not in Singleton.singleton_map:
                return super().__call__(*args, **kwargs)
            else:
                raise RuntimeError(f"Singleton {cls.__name__} already initialized. Use {cls.__name__}(init=False) to get the instance.")
        else:
            if cls.__name__ in Singleton.singleton_map:
                return cls.get_by_name(cls.__name__)
            else:
                raise RuntimeError(f"Singleton {cls.__name__} not initialized. Call {cls.__name__}() first.")

    @classmethod
    def get_all(cls) -> dict[str, Any]:
        return cls.singleton_map

    @classmethod
    def get_by_name(cls, name: str) -> Any:
        current = cls.get_all().get(name, None)
        if current is None:
            raise RuntimeError(f"Singleton {name} not found.")
        return current

    @classmethod
    def get_by_type(cls, t: type | str) -> Any:
        if isinstance(t, str):
            searching_name = t
        elif isinstance(t, type):
            searching_name = t.__name__
        else:
            raise TypeError(f"Type of 't' must be 'str' or 'type', not '{type(t)}'.")
        for _, instance in cls.get_all().items():
            if isinstance(t, str):
                if instance.__class__.__name__ == searching_name:
                    return instance

                # get all bases
                bases = instance.__class__.__bases__
                for base in bases:
                    if base.__name__ == searching_name:
                        return instance
            elif isinstance(t, type):
                if isinstance(instance, t):
                    return instance
        raise RuntimeError(f"Singleton {searching_name} not found.")
